@@ -2,10 +2,16 @@
 # coding: utf-8
 import numpy as np
 import matplotlib.pyplot as plt
-from progress.bar import Bar
 from tqdm import tqdm 
 from joblib import Parallel, delayed,cpu_count  
+from joblib.externals.loky import get_reusable_executor
 
+
+def lnlike(data,model,error):
+  # default log-likelihood function
+  chi2=np.nansum((data - model)**2 / error**2)      
+  return -0.5*chi2
+      
 def unwrap_self(args, **kwarg):
     return gastimator.run_a_chain(*args, **kwarg)
 
@@ -34,16 +40,27 @@ class gastimator:
       self.lastchainll=None
       self.lastchainaccept=None
       self.nprocesses=np.int(cpu_count())-1
-
+      self.lnlike_func=lnlike
+      
+    
+  # def likelihood(self,values):
+  #
+  #   priorval=1
+  #   for ival,prior in enumerate(self.prior_func):
+  #       if callable(prior):
+  #           priorval*=prior(values[ival],allvalues=values,ival=ival)
+  #
+  #   model=self.model(values,*self.args,**self.kwargs)
+  #   return self.lnlike_func(self.fdata,model,self.error) + np.log(priorval, where=(priorval!=0))
   def likelihood(self,values):
     
-    priorval=1
+    priorval=0
     for ival,prior in enumerate(self.prior_func):
         if callable(prior):
-            priorval*=prior(values[ival],allvalues=values,ival=ival)
+            priorval+=prior(values[ival],allvalues=values,ival=ival)
     
-    chi2=np.nansum((self.fdata - self.model(values,*self.args,**self.kwargs))**2 / self.error**2)
-    return -0.5*chi2 + np.log(priorval, where=(priorval!=0))
+    model=self.model(values,*self.args,**self.kwargs)
+    return self.lnlike_func(self.fdata,model,self.error) + priorval 
 
     
 
@@ -93,13 +110,13 @@ class gastimator:
   
   
 
-  def chain(self, values, niter, knob, plot=True, holdknob=False, fig=None, ax=None):
+  def chain(self, values, niter, knob, plot=True, holdknob=False, fig=None, ax=None,progress=False,progid=0):
      
 
       if plot:
           fig, ax=self.make_plots(niter)
           plt.pause(0.00001)
-          
+      
       gibbs_change=self.rng.randint(0,self.change.size,niter)
 
       outputvals=np.empty((self.npars,niter))
@@ -113,17 +130,21 @@ class gastimator:
       outputll[0]=bestll
       n_accept=0.
       
+      if progress and (progid==0):
+          pbar = tqdm(total=(niter)*self.nprocesses)
+
       for i in range(1,niter):
           newval, newll, accept, tryvals, knob =self.evaluate_step(outputvals[:,i-1],outputll[i-1],self.change[gibbs_change[i]],knob,holdknob=holdknob)
-          if holdknob and not self.silent and self.nprocesses == 1: self.bar.next()
+          if progress and (progid==0): 
+              pbar.update(self.nprocesses)
           if plot:
               self.update_plot(fig, ax,self.change[gibbs_change[i]],i,tryvals, accept)
-
-              
           outputvals[:,i]=newval
           outputll[i]=newll
           accepted[i]=accept
           n_accept+=accept
+          
+          
      
       acceptrate=np.float(n_accept)/niter
       
@@ -138,6 +159,8 @@ class gastimator:
          outputll=outputll[accepted.astype(bool)]
       
       if plot: plt.close()
+      if progress and (progid==0):
+          pbar.close()
       return outputvals, outputll, accepted, acceptrate, knob
 
   def factors(self,n):  
@@ -175,18 +198,20 @@ class gastimator:
 
 
         
-  def run_a_chain(self,startguess,niters,numatonce,knob,plot=True,final=False):
+  def run_a_chain(self,startguess,niters,numatonce,knob,plot=True,final=False,progid=0):
         count=0
         converged=False
         oldmean=self.guesses*0.0
         newmean=0.0
         startpoint=startguess.copy()
-        
+        self.rng=np.random.RandomState()
         if not final:
             while (count < niters) and (not converged):
 
-                outputvals, outputll, accepted, acceptrate, knob = self.chain(startpoint, numatonce, knob, plot=plot, holdknob=False)
-                
+                outputvals, outputll, accepted, acceptrate, knob = self.chain(startpoint, numatonce, knob, plot=plot, holdknob=False,progress=False)
+
+                    
+                    
                 if acceptrate*numatonce > 1:
                     w,=np.where(accepted) 
                     startpoint=outputvals[:,w[-1]]
@@ -197,35 +222,34 @@ class gastimator:
                     self.lastchain=outputvals
                     self.lastchainll=outputll
                     self.lastchainaccept=accepted
-                    if (not self.silent): print("     Chain has not converged - Accept rate: "+str(acceptrate))
+                    if (not self.silent): print("     Chain "+str(progid)+" has not converged - Accept rate: "+str(acceptrate))
                 else:
                     self.lastchain=np.append(self.lastchain,outputvals,axis=1)
                     self.lastchainll=np.append(self.lastchainll,outputll)
                     self.lastchainaccept=np.append(self.lastchainaccept,accepted)
                     test=(np.abs(newmean-oldmean) < self.precision)
-                    if test.sum() == test.size and (acceptrate >= self.targetrate):
+                    if test[self.change].sum() == test[self.change].size and (acceptrate >= self.targetrate):
                          converged=1      
-                         if (not self.silent): print("Chain converged: LL: "+str(np.max(outputll))+" - Accept rate:"+str(acceptrate))
+                         if (not self.silent): print("Chain "+str(progid)+" converged: LL: "+str(np.max(outputll))+" - Accept rate:"+str(acceptrate))
                     else:
-                         if  (not self.silent): print("     Chain has not converged - Accept rate: "+str(acceptrate))
-                         if test.sum() == test.size:
-                             if  (not self.silent): print("     Target rate not reached")
+                         if  (not self.silent): print("     Chain "+str(progid)+" has not converged - Accept rate: "+str(acceptrate))
+                         if test[self.change].sum() == test[self.change].size:
+                             if  (not self.silent): print("     --> Target rate not reached")
                          else: 
-                             if  (not self.silent): print("     Still varying: "+str(self.labels[~test]))
+                             if  (not self.silent): print("     --> Still varying: "+str(self.labels[self.change][~test[self.change]]))
                      
                 oldmean=newmean    
                 count += numatonce
-            if not converged: print('WARNING: Chain did not converge in '+str(niters)+' steps')
+            if not converged: print('WARNING: Chain '+str(progid)+' did not converge in '+str(niters)+' steps')
            
                              
         else:
-            if (self.silent == False)&(self.nprocesses==1):
-                with Bar('Final chain', max=niters-1, suffix='%(percent)d%%') as self.bar:
-                    outputvals, outputll, accepted, acceptrate, knob = self.chain(startpoint, niters, knob, plot=False, holdknob=True)
-            else:
-                self.rng=np.random.RandomState() #refresh the RNG
-                outputvals, outputll, accepted, acceptrate, knob = self.chain(startpoint, niters, knob, plot=False, holdknob=True)
-           
+            #if (self.silent == False)&(self.nprocesses==1):
+            #        outputvals, outputll, accepted, acceptrate, knob = self.chain(startpoint, niters, knob, plot=False, holdknob=True,progress=True,progid=progid)
+            #else:
+            #self.rng=np.random.RandomState() #refresh the RNG
+            outputvals, outputll, accepted, acceptrate, knob = self.chain(startpoint, niters, knob, plot=False, holdknob=True,progress=True,progid=progid)
+            
         best_knob=knob
         return outputvals, outputll, best_knob
                     
@@ -236,10 +260,27 @@ class gastimator:
     self.precision=np.array(self.precision)
     self.labels=np.array(self.labels)
     
+                
     if np.any(self.guesses == None):
          raise Exception('Please set initial guesses')
     self.npars=self.guesses.size
     
+    if np.all(np.array(self.prior_func) == None):
+             self.prior_func=(np.zeros(self.npars, dtype=bool)).tolist()
+    else:
+        try:
+            if len(self.prior_func) != self.npars:
+                 raise Exception('Number of priors given does not match number of parameters')
+        except:
+            if self.npars != 1:
+                raise Exception('Number of priors given does not match number of parameters')
+        
+
+
+             
+    
+    
+    
     names=["minimum","maximum","precision","labels"]    
     check=[self.min,self.max,self.precision,self.labels]
 
@@ -254,22 +295,29 @@ class gastimator:
              self.fixed=np.zeros(self.npars, dtype=bool)
              print("You did not specify if any variables are fixed - I will continue assuming that none are")
         
-    if np.any(self.prior_func == None):
-             self.prior_func=np.zeros(self.npars, dtype=bool)    
+
              
     if np.any((self.max-self.min) < 0):
              raise Exception('Parameter(s) '+str(self.labels[(self.max-self.min) < 0])+' have incorrect minumum/maximum bounds')
  
     if np.any((self.guesses<self.min)):
-             raise Exception('Parameter(s) '+str(self.labels[(self.guesses<self.min)])+' have an initial guess lower than the minimum allowed.')
+             raise Exception('Parameter(s) '+str(self.labels[(self.guesses<self.min)])+' have an initial guess ('+str(self.guesses[(self.guesses<self.min)])+') lower than the minimum allowed ('+str(self.min[(self.guesses<self.min)])+').')
              
     if np.any((self.guesses>self.max)):
-             raise Exception('Parameter(s) '+str(self.labels[(self.guesses>self.max)])+' have an initial guess higher than the maximum allowed.')
+             raise Exception('Parameter(s) '+str(self.labels[(self.guesses>self.max)])+' have an initial guess higher ('+str(self.guesses[(self.guesses>self.max)])+') than the maximum allowed ('+str(self.max[(self.guesses>self.max)])+').')
 
 
   def run(self, fdata, error, niters, numatonce=None, burn=None, nchains=1, plot=True, output=None, seed=None): 
     # check all required inputs set
     self.input_checks()
+    
+    if not self.silent:
+        stri=''
+        for name,pr in zip(self.labels,self.prior_func):
+            if callable(pr):
+                stri=stri+' '+name
+        if stri!='':
+            print("Non uniform priors used: "+stri)
 
     # set up variables needed
     self.fdata=fdata
@@ -290,17 +338,35 @@ class gastimator:
     else:
         self.burn=burn
 
+    if (self.silent==False)&(self.nprocesses>1):
+        verboselev=10
+    else:
+        verboselev=0
 
+    self.progline=[]
     for chainno in range(0,nchains):
-        if not self.silent: print('Doing chain '+np.str(chainno+1))
-        knob=(0.5*(self.max-self.min))
-        outputvals, outputll, best_knob = self.run_a_chain(self.guesses,niters,numatonce,knob,plot=plot)
-        if (np.max(outputll) > verybestll) or (chainno == 0):
-            if not self.silent: print("Best chain so far!")
-            w,=np.where(outputll == np.max(outputll))
-            verybestvalues=outputvals[:,w[0]].reshape(self.npars) 
-            verybestknob=best_knob
-            verybestll=np.max(outputll)
+        self.progline.append('Doing chain '+np.str(chainno+1))
+        
+    knob=(0.5*(self.max-self.min))
+    
+    
+        
+    ## do initial chains in parallel    
+    par= Parallel(n_jobs= self.nprocesses, verbose=verboselev)
+    results=par(delayed(unwrap_self)(i) for i in zip([self]*nchains, [self.guesses]*nchains,[int(float(niters))]*nchains,[numatonce]*nchains,[knob]*nchains, [False]*nchains, [False]*nchains,np.arange(nchains)))
+    results=np.array(results,dtype=object)
+    par._terminate_backend()
+    get_reusable_executor().shutdown(wait=True)
+    ##
+    #debug line #results=unwrap_self((self, self.guesses,int(float(niters)),numatonce,knob, False, False,0))
+
+    bestchain=np.argmax(np.max(np.stack(results[:,1]),axis=1))
+    bestvalinbestchain=np.argmax(results[bestchain,1])
+    
+
+    verybestvalues=np.stack(results[bestchain,0])[:,bestvalinbestchain]
+    verybestknob=results[bestchain,2]
+    verybestll=np.stack(results[bestchain,1])[bestvalinbestchain]
             
             
     if not self.silent: 
@@ -309,13 +375,15 @@ class gastimator:
             print("  "+self.labels[i]+":",verybestvalues[i])
         print("Starting final chain")
 
-    if (self.silent==False)&(self.nprocesses>1):
-        verboselev=10
-    else:
-        verboselev=0
+
     results = []
-    results= Parallel(n_jobs= self.nprocesses, verbose=verboselev)(delayed(unwrap_self)(i) for i in zip([self]*self.nprocesses, [verybestvalues]*self.nprocesses,[int(float(niters)/float(self.nprocesses))]*self.nprocesses,[numatonce]*self.nprocesses,[verybestknob]*self.nprocesses, [False]*self.nprocesses, [True]*self.nprocesses))
+    par= Parallel(n_jobs= self.nprocesses, verbose=verboselev)
+    
+    results=par(delayed(unwrap_self)(i) for i in zip([self]*self.nprocesses, [verybestvalues]*self.nprocesses,[int(float(niters)/float(self.nprocesses))]*self.nprocesses,[numatonce]*self.nprocesses,[verybestknob]*self.nprocesses, [False]*self.nprocesses, [True]*self.nprocesses,np.arange(self.nprocesses)))
     results=np.array(results,dtype=object)
+    par._terminate_backend()
+    get_reusable_executor().shutdown(wait=True)
+    
     outputvalue= np.concatenate(results[:,0],axis=1)
     outputll= np.concatenate(results[:,1])
     
